module Control.HM.Encapsulate where

import Data.HListExtras
import Control.Monad
import Control.Monad.Identity
import Pointless.Combinators



data LEFT a = LEFT a
instance Show a => Show (LEFT a) where show (LEFT a) = "LEFT " ++ show a

data RIGHT a = RIGHT a
instance Show a => Show (RIGHT a) where show (RIGHT a) = "RIGHT " ++ show a

data HEither e l = HLeft e | HRight l deriving (Eq)
data HVoid 

instance ( Show l, Show e ) => Show (HEither e l)
  where
    show (HLeft e) = show e
    show (HRight l) = show l

instance Show HVoid where show _ = "HVoid"

type (:++:) e l = HEither e l

infixr 2 :++:

class Sum l s x | l s -> x where
  select :: l -> s -> Maybe x
  inject :: l -> x -> s

instance (TypeEq l l' b, Sum' b l (HEither (l', x) xs) x' ) 
  => 
    Sum l (HEither (l', x) xs) x' 
    where
    select l s = select' b l s 
        where b = typeEq l (_L::l')
    inject l x = inject' b l x
        where b = typeEq l (_L::l') 


class Sum' b l s x | b l s -> x where
  select' :: b -> l -> s -> Maybe x
  inject' :: b -> l -> x -> s


instance Sum' HTrue l (HEither (l, x) xs) x 
    where
    select' _ _ (HLeft (_, x)) = Just x
    inject' _ l x = HLeft (l, x)

instance ( Sum l ys y ) => Sum' HFalse l (HEither lx ys) y 
    where
    select' b l (HRight ys) = select l ys
    inject' b l y = HRight (inject l y)



class Encapsulate cp is st os m | cp -> is st os m 
  where
    (->>) :: cp -> (st, is) -> m (st, os)
 

instance ( Encapsulate (st -> fs) is st os m, Monad m ) 
  => 
   Encapsulate (st -> HCons (l', e -> m (st, r)) fs)
                (HEither (l', e) is)
                st
                (HEither (l', r) os)
                m
  where
    (->>) g (st, HLeft (_, e)) = do 
                                   let (HCons (l, f) _) = g st
                                   (st', r) <- f e
                                   return (st', HLeft (l, r))
    (->>) fs (st, HRight is) = do
                                 let (HCons _ fs') = fs st
                                 (st', os) <- (->>) ((\st -> fs') :: st -> fs) (st, is)
                                 return (st', HRight os)
                                        


instance Monad m => Encapsulate (st -> HCons (l', e -> m (st, r)) HNil)
                                      (HEither (l', e) HVoid) 
                                      st 
                                      (HEither (l', r) HVoid)
                                      m 
  where
    (->>) g = \(st, HLeft (_, e)) -> do 
                                       let (HCons (l, f) _) = g st
                                       (st', r) <- f e
                                       return (st', HLeft (l, r))


type CpTL s l i o m = s -> l -> i -> m (s, (l, o))





