module Control.HM.Tensors where


import Control.HM.Encapsulate

import Pointless.Combinators
import HListPrelude
import Data.HListExtras
import Control.Monad.State


-- ***************************************************** APPLICATION ***********************
class Apl it o1 l i st o m | l it o1 -> i o
  where
    (@.) :: ((st, it) -> (m (st, o1))) -> st -> l -> i -> m (st, (l, o)) --CpTL st l i o m

instance ( Monad m, Sum l o1 o, Sum l it i ) => Apl it o1 l i st o m
  where 
    (@.) cp st l i = do 
                        let input = inject l i :: it
                        (st', output) <- cp (st, input)
                        let (Just output') = select l output
                        return (st', (l, output'))




-- ***************************************************** SIMPLE APPLICATION ********************
class Apls i st o m 
  where
    (.@) :: ((st, i) -> (m (st, o))) -> st -> i -> m (st, o)

instance Apls i st o m
  where
    (.@) cp s i = cp (s, i) 




-- *********************************************************** DoCompG ************************
class DoCompG i st o m 
  where 
    doCompG :: ((st, i) -> (m (st, o))) -> st -> [i] -> StateT [o] m st

instance Monad m => DoCompG i st o m 
  where
    doCompG cp st [input] = do
                                (st', output) <- lift $ (.@) cp st input
                                s <- get 
                                put $ s++[output]
                                return st'
    doCompG cp st (input:is) = do
                                (st', output) <- lift $ (.@) cp st input
                                s <- get
                                put $ s++[output]
                                doCompG cp st' is





-- *********************************************************** DoCompIO ************************

class ToIO m => DoCompIO it o1 st o m 
  where 
    doCompIO :: ((st, it) -> m (st, o1)) -> StateT st IO o

instance ( Read it, Show o1, Monad m, MonadIO m, ToIO m ) => DoCompIO it o1 st o m
  where
    doCompIO cp = do
                        lift $ putStr "\nAction: "
                        i <- lift getLine
                        st <- get
                        (st', output) <- lift $ toIO $ (.@) cp st (read i)
                        lift $ putStrLn $ show output
                        put st'
                        doCompIO cp 


class Monad m => ToIO m
    where   
        toIO :: m a -> IO a

instance ToIO IO
    where
        toIO = id

instance ToIO Maybe 
    where
        toIO Nothing = error "Nothing"
        toIO (Just x) = return x

instance ToIO []
    where
        toIO = return.head -- not good





-- ************************************************ PARALLEL *****************************

class Parallel s1 l i o l1 s2 l' i' o' l2 m cp3 | s1 l i o l1 s2 l' i' o' l2 m -> cp3
  where 
    (|><|) :: (s1 -> HCons (l, i -> m (s1, o)) l1) -> (s2 -> HCons (l', i' -> m (s2, o')) l2) -> cp3

instance ( ToPairs (HCons (l, i -> m (s1, o)) l1)
                   (HCons (l', i' -> m (s2, o')) l2) 
                   (s1, s2)
                   (HCons ((l, l'), (i, i') -> m ((s1, s2), (o, o'))) lstf) )
  => 
    Parallel s1 l i o l1 s2 l' i' o' l2 m 
             ((s1, s2) -> HCons ((l, l'), (i, i') -> m ((s1, s2), (o, o'))) lstf)
  where
    cp1 |><| cp2 = \(s1, s2) -> toPairs (cp1 s1) (cp2 s2) (s1, s2)


class ToPairs l1 l2 s l | l1 l2 s -> l 
  where 
    toPairs :: l1 -> l2 -> s -> l

instance ToPairs  HNil l s HNil 
  where 
    toPairs _ _ _ = HNil


instance ( ToPairs ls (HCons (l', i' -> m (s2, o')) ls') (s1, s2) rl, Monad m,
           ToPair (l, i -> m (s1, o)) (HCons (l', i' -> m (s2, o')) ls') (s1, s2) 
                  (HCons ((l, l'), (i, i') -> m ((s1, s2), (o, o'))) lr'), 
           HAppend (HCons ((l, l'), (i, i') -> m ((s1, s2), (o, o'))) lr') rl
                   (HCons ((l, l'), (i, i') -> m ((s1, s2), (o, o'))) lsf) ) 
  => 
    ToPairs (HCons (l, i -> m (s1, o)) ls) 
            (HCons (l', i' -> m (s2, o')) ls') 
            (s1, s2) 
            (HCons ((l, l'), (i, i') -> m ((s1, s2), (o, o'))) lsf)
  where
    toPairs (HCons (l, f) ls) l2 (s1, s2) = toPair (l, f) l2 (s1, s2) `hAppend`
                                            toPairs ls l2 (s1, s2)


class ToPair p l2 s lf | p l2 s -> lf 
  where 
    toPair :: p -> l2 -> s -> lf

instance ToPair p HNil s HNil 
  where 
    toPair _ _ _ = HNil

instance ( ToPair (l, i -> m (s1, o)) ls (s1, s2) ls', Monad m ) 
  => 
    ToPair (l, i -> m (s1, o)) (HCons (l', i' -> m (s2, o')) ls) (s1, s2) 
           (HCons ((l, l'), (i, i') -> m ((s1, s2), (o, o'))) ls')
  where
    toPair (l, f) (HCons (l', f') ls) (s1, s2) = 
     HCons ((l, l'), \(i, i') -> do {(s1', o) <- f i; (s2', o') <- f' i'; 
                                     return ((s1', s2'), (o, o'))}) 
           (toPair (l, f) ls (s1, s2))


class ToRight l l' | l -> l', l' -> l 

instance ToRight HVoid HVoid 

instance ToRight ls ls' => ToRight (HEither (l, i) ls) (HEither (RIGHT l, i) ls')





-- ********************************************** CHOICE  ***********************************

class Choice s1 l i o l1 s2 l' i' o' l2 m cp3 | s1 l i o l1 s2 l' i' o' l2 m -> cp3
  where 
    (|+|) :: (s1 -> HCons (l, i -> m (s1, o)) l1) -> (s2 -> HCons (l', i' -> m (s2, o')) l2) -> cp3

instance ( ToLeftLst (HCons (l, i -> m (s1, o)) l1) (s1, s2) 
                     (HCons (LEFT l, i -> m ((s1, s2), o)) l1f), 
           
           ToRightLst (HCons (l', i' -> m (s2, o')) l2) (s1, s2) 
                      (HCons (RIGHT l', i' -> m ((s1, s2), o')) l2f), 

            LeftE (HCons (LEFT l, i -> m ((s1, s2), o)) l1f) i' o'
                  (HCons (LEFT l, Either i i' -> m ((s1, s2), Either o o')) rl''),

            RightE (HCons (RIGHT l', i' -> m ((s1, s2), o')) l2f) i o
                   (HCons (RIGHT l', Either i i' -> m ((s1, s2), Either o o')) rl2''),

            HAppend (HCons (LEFT l, Either i i' -> m ((s1, s2), Either o o')) rl'') 
                    (HCons (RIGHT l', Either i i' -> m ((s1, s2), Either o o')) rl2'') 
                    (HCons (LEFT l, Either i i' -> m ((s1, s2), Either o o')) lstf), 
                      Monad m )
  => 
    Choice s1 l i o l1 s2 l' i' o' l2 m 
            ((s1, s2) -> (HCons (LEFT l, Either i i' -> m ((s1, s2), Either o o')) lstf))
  where
    cp1 |+| cp2 = \(s1, s2) -> (leftE (toLeftLst (cp1 s1) (s1, s2)) (_L::i') (_L::o'))
                                         `hAppend` 
                                 (rightE (toRightLst (cp2 s2) (s1, s2)) (_L::i) (_L::o))


class LeftE cp i' o' cp' | cp i' o' -> cp'
    where
        leftE :: cp -> i' -> o' -> cp'

instance LeftE HNil i' o' HNil
    where
        leftE _ _ _ = HNil

instance ( LeftE rl i' o' rl', Monad m )
    =>
        LeftE (HCons (l, i -> m (s, o)) rl) i' o' 
              (HCons (l, Either i i' -> m (s, Either o o')) rl')
    where
        leftE (HCons (l, f) rl) i' o' = 
            HCons (l, \(Left i) -> f i >>= \(s, o) -> return (s, Left o)) 
                  (leftE rl i' o')


class RightE cp i' o' cp' | cp i' o' -> cp'
    where
        rightE :: cp -> i' -> o' -> cp'

instance RightE HNil i' o' HNil
    where
        rightE _ _ _ = HNil

instance ( RightE rl i' o' rl', Monad m )
    =>
        RightE (HCons (l, i -> m (s, o)) rl) i' o' 
               (HCons (l, Either i' i -> m (s, Either o' o)) rl')
    where
        rightE (HCons (l, f) rl) i' o' = 
            HCons (l, \(Right i) -> f i >>= \(s, o) -> return (s, Right o)) 
                  (rightE rl i' o')


class ToLeftLst l s l' | l s -> l', l' s -> l 
  where 
    toLeftLst :: l -> s -> l'

instance ToLeftLst  HNil s HNil 
  where 
    toLeftLst _ _ = HNil

instance ( ToLeftLst ls (s1, s2) ls', Monad m ) 
  => 
    ToLeftLst (HCons (l, i -> m (s1, o)) ls) (s1, s2) 
              (HCons (LEFT l, i -> m ((s1, s2), o)) ls')
  where
    toLeftLst (HCons (l, f) ls) (s1, s2) = 
             HCons (_L::LEFT l, \i -> do {(s1', o) <- f i; return ((s1', s2), o)})
                   (toLeftLst ls (s1, s2))

class ToRightLst l s l' | l s -> l', l' s -> l 
  where 
    toRightLst :: l -> s -> l'

instance ToRightLst HNil s HNil 
  where 
    toRightLst _ _ = HNil

instance ( ToRightLst ls (s1, s2) ls', Monad m) 
  => 
    ToRightLst (HCons (l, i -> m (s2, o)) ls) (s1, s2) 
               (HCons (RIGHT l, i -> m ((s1, s2), o)) ls')
  where
    toRightLst (HCons (l, f) ls) (s1, s2) =
          HCons (_L::RIGHT l, \i -> do {(s2', o) <- f i; return ((s1,s2'), o)})
                (toRightLst ls (s1, s2))





-- ******************************************* HOOK ************************************
class Hook cp l cp' | cp l -> cp' 
  where 
    hook :: cp -> l -> cp'

instance ( HAppend lf1 lf2 lf', ConstH lf s l lf1, HSnd HNil l lfr, DeleteMany lfr lf lf2 )
  => 
    Hook (s -> lf) l (s -> lf')
  where
    hook f l = \s -> 
                     let fs = f s
                         lfr = hsnd HNil l
                         cons = constH f s l
                         fs' = deleteMany lfr fs
                     in cons `hAppend` fs'


class HSnd ac l ac' | ac l -> ac'
    where
        hsnd :: ac -> l -> ac'

instance HSnd ac HNil ac
    where
        hsnd ac _ = ac

instance (  HSnd' b1 ac l1 lf1, HSnd' b2 lf1 l2 lf3, HSnd lf3 rl rl',
            HTMember l1 ac b1, HTMember l2 lf1 b2 )
    => 
        HSnd ac (HCons (l, (l1, l2)) rl) rl'
    where
        hsnd ac (HCons (l, (l1, l2)) rl) = hsnd (hsnd' b2 (hsnd' b1 ac l1) l2) rl

          where
            b1 = hTMember l1 ac
            b2 = hTMember l2 (_L :: lf1)


class HSnd' b ac l ac' | b ac l -> ac'
    where
        hsnd' :: b -> ac -> l -> ac'

instance HSnd' HTrue ac l ac
    where
        hsnd' _ ac _ = ac

instance HSnd' HFalse ac l (HCons l ac)
    where
        hsnd' _ ac l = HCons l ac




class ConstH lf s l lf' | lf s l -> lf' 
  where 
    constH :: (s -> lf) -> s -> l -> lf'

instance ConstH lf s HNil HNil 
  where 
    constH _ _ _ = HNil

instance ( ConstH lf s rl y, ConstF lf s ll f ) 
  =>
    ConstH lf s (HCons (l, ll) rl) (HCons (l, f) y)
  where
    constH lf s (HCons (l, ll) rl) = let f = constF lf s ll 
                                     in HCons (l, f) (constH lf s rl)


class ConstF lf s ll f | lf s ll -> f 
  where 
    constF :: (s -> lf) -> s -> ll -> f


instance ( Monad m, 
           HasFieldM l lf (Either i z -> m (s, Either o z)), 
           HasFieldM l' lf (Either i z -> m (s, Either o z)) ) 
  => 
    ConstF lf s (l, l') (Either i z -> m (s, Either o z))
  where
    constF lf s (l, l') = \i -> do 
                                                let f1 = selct l (lf s) 
                                                (s', o) <- f1 i
                                                let f2 = selct l' (lf s')
                                                case o of
                                                    (Right z) -> f2 (Right z)
                                                    (Left o') -> return (s', o)
                                                




-- *************************************************** WRAP **********************************

class Wrap cp f g cp' | cp f g -> cp'
    where
        wrap :: cp -> f -> g -> cp'


instance (  Monad m, Wrap' (HCons (l, i -> m (s, o)) r) (i' -> i) (o -> o') 
                                 (HCons (l, i' -> m (s, o')) rt) )
    => 
        Wrap (s -> HCons (l, i -> m (s, o)) r) (i' -> i) (o -> o') 
             (s -> HCons (l, i' -> m (s, o')) rt)
    where
        wrap cp f g = \s -> wrap' (cp s) f g 


class Wrap' lf f g lf' | lf f g -> lf'
    where
        wrap' :: lf -> f -> g -> lf'


instance Wrap' HNil f g HNil
    where
        wrap' _ _ _ = HNil


instance ( Monad m, Wrap' r (i' -> i) (o -> o') rt )
    => 
        Wrap' (HCons (l, i -> m (s, o)) r) (i' -> i) (o -> o') 
              (HCons (l, i' -> m (s, o')) rt)
    where
        wrap' (HCons (l, f1) r) f g = 
                                  HCons (l, \i' -> do {(s, o) <- f1 (f i'); return (s, g o)}) 
                                        (wrap' r f g)






-- *************************************************** REFACT *****************************

class Refact cp l cp' | cp l -> cp'
  where 
    refact :: cp -> l -> cp'

instance Refact (s -> lf) HNil (s -> lf) 
  where 
    refact f _ = f

instance ( Constl lf l' ct, Remov l tf ftf, HAppend lf ct tf ) 
  => 
    Refact (s -> lf) (HCons l l') (s -> ftf)
  where
    refact f (HCons l l') = \s -> let fs = f s
                                      ct = constl fs l' 
                                  in remov l (fs `hAppend` ct)

instance ( Constl (HCons (h, f') rp) (HCons (a, b) l) (HCons cf rf), 
           HAppend (HCons (h, f') rp) (HCons cf rf) (HCons (h, f') l''),
           HAppend rp (HCons cf rf) l'', HList rp )
  => 
    Refact (s -> HCons (h, f') rp) (HCons HNil (HCons (a, b) l)) (s -> HCons (h, f') l'')
  where
    refact f (HCons HNil (HCons (a, b) l)) = \s -> let lf = f s
                                                 in lf `hAppend` constl lf (HCons (a, b) l)

instance Remov rl lf tf => Refact (s -> lf) (HCons rl HNil) (s -> tf)
  where
    refact f (HCons rl _) = \s -> remov rl (f s) 


class Remov l l' l'' | l l' -> l''
  where 
    remov :: l -> l' -> l''

instance Remov HNil l l 
  where 
    remov _ = id


instance ( Remov ls ls' r, Remov1 b l l' ls', HHead l' (h, f'), TypeEq l h b ) 
  => 
    Remov (HCons l ls) l' r
  where
    remov (HCons l ls) l' = let h = fst $ hHead l'
                            in remov ls (remov1 (typeEq l h) l l') 


class Remov1 b l ls ls' | b l ls -> ls' 
  where 
    remov1 :: b -> l -> ls -> ls'

instance Remov1 HFalse l (HCons (l', f) HNil) (HCons (l', f) HNil) 
  where 
    remov1 _ _ = id

instance Remov1 b l HNil HNil
  where
    remov1 _ _ _ = HNil

instance Remov1 HTrue l (HCons (l', f) ls) ls
  where
    remov1 _ l (HCons _ ls) = ls 
                             
    
instance ( Remov1 b l ls ls', HHead ls (h, f'), TypeEq l h b ) 
  => 
    Remov1 HFalse l (HCons (l', f) ls) (HCons (l', f) ls')
  where
    remov1 _ l (HCons p ls) = let h = fst $ hHead ls 
                              in HCons p (remov1 (typeEq l h) l ls)


class Constl lg l l' | lg l -> l' 
  where 
    constl :: lg -> l -> l'

instance Constl lg HNil HNil 
  where 
    constl _ _ = HNil

instance ( Constl lg rll rf', HasFieldM l' lg f, ConstC f ll c, HAppend c rf' (HCons cf rf) ) 
  => 
    Constl lg (HCons (l', ll) rll) (HCons cf rf)
  where
    constl lg (HCons (l', ll) rll) = let f = selct l' lg 
                                     in constC f ll `hAppend` constl lg rll

selct l r = hLookupByLabelM l r
selct2 l r = hLookupByLabel l r


class ConstC f ll f' | f ll -> f', f' -> ll 
  where 
    constC :: f -> ll -> f'

instance ConstC f HNil HNil 
  where 
    constC _ _ = HNil

instance ConstC f rl rlf => ConstC f (HCons l rl) (HCons (l, f) rlf)
  where
    constC f (HCons l rl) = HCons (l, f) (constC f rl)







-- *************************************************** LIFT *****************************
class Lift i o s l m cp | i o s l m -> cp
  where
    cpLift :: (i -> o) -> l -> (s -> HCons (l, i -> m (s, o)) HNil)

instance Monad m
  =>
    Lift i o s l m (s -> HCons (l, i -> m (s, o)) HNil)
  where
    cpLift f l = \s -> HCons (l, \i -> return (s, f i)) HNil





-- *************************************************** COMPOSITION *****************************

class Composition cp1 cp2 cp3 | cp1 cp2 -> cp3
  where
    (|.|) :: cp1 -> cp2 -> cp3

instance ( ComposeAll lf1 lf2 lf )
  =>
    Composition (s1 -> lf1) (s2 -> lf2) ((s1, s2) -> lf)
  where
    cp1 |.| cp2 = \(s1, s2) -> composeAll (cp1 s1) (cp2 s2)


class ComposeAll lf1 lf2 lf | lf1 lf2 -> lf
  where
    composeAll :: lf1 -> lf2 -> lf

instance ComposeAll HNil lf2 HNil
  where
    composeAll _ _ = HNil

instance (ComposeOne l f lf2 lf1', ComposeAll lf1 lf2 lf2', HAppend lf1' lf2' lf)
  =>
    ComposeAll (HCons (l, f) lf1) lf2 lf
  where
    composeAll (HCons (l, f) lf1) lf2 = composeOne l f lf2 `hAppend` composeAll lf1 lf2


class ComposeOne l f lf2 lf | l f lf2 -> lf
  where
    composeOne :: l -> f -> lf2 -> lf

instance ComposeOne l f HNil HNil
  where
    composeOne _ _ _ = HNil

instance (Monad m, ComposeOne l (i -> m (s1, o)) rl lf)
  =>
    ComposeOne l (i -> m (s1, o)) (HCons (l', o -> m (s2, o')) rl) 
                                  (HCons ((l, l'), i -> m ((s1, s2), o')) lf)
  where
    composeOne l f (HCons (l', g) rl) = 
        HCons ((l,l'), \i -> do {(s1, o) <- f i; (s2, o') <- g o; return ((s1, s2), o')}) 
              (composeOne l f rl)

