-----------------------------------------------------------------------------
-- |
-- Maintainer  :  joost.visser@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- Some utility functions for unit testing.
--
-----------------------------------------------------------------------------

module Debug.HUnitExtras where

import HUnit
import Data.Unique (newUnique, hashUnique)
import SGLR (sglr, exitOK)
import ATermLib
import System
import System.IO

-----------------------------------------------------------------------------

-- | Run a list of tests.
runTestsTT :: [Test] -> IO Counts
runTestsTT = runTestTT . TestList

-- | Create a labeled test.
labeledTest :: String -> Assertion -> Test
labeledTest l t
  = TestLabel l (TestCase t)


-- | Represent a test-suite
testSuite :: String -> [Test] -> Test
testSuite s st
   = TestLabel s (TestList st)



-- | Parses a string with sglr. It requires a string to parse, a start symbol
--   of the grammar, and a string with the parse table path.
parseString :: ATermConvertible a => String -> String -> String -> IO a 
parseString string startSymbol sdfTable
   = viaTmpFile string $ \f -> do sglr sdfTable f startSymbol


-- | Perform given IO action after creating a temporary file with the
--   given string as content
viaTmpFile :: String -> (FilePath -> IO a) -> IO a
viaTmpFile content action = do
  unique <- newUnique
  let fileName = "/tmp/hunit"++(show . hashUnique $ unique)++".tmp"
  --putStrLn $ "\nOpening file: "++fileName
  writeFile fileName content
  --putStrLn $ "Wrote to file: "++fileName++": "++content
  a <- action fileName
  exitOK $ system $ "rm "++fileName
  --putStrLn $ "Removed file: "++fileName
  return a

-- | Assert the succes of a given IO computation. If it fails, 
--   the given String is used in the HUnit report.
assertSuccess :: String -> IO a -> Assertion
assertSuccess msg m = do
  (m >> return ()) `catch` \_ -> assertFailure msg

-----------------------------------------------------------------------------
