{-# OPTIONS -fglasgow-exts #-}

module Pointless.Examples.PwPf where

import Pointless.Combinators
import Pointless.Combinators.Uncurried
import Pointless.Functors
import Pointless.RecursionPatterns
import Pointless.Isomorphisms hiding (coswap)

-- Some macros

unpnt f = app . (f . bang /\ id)
casesum e l r = app . (eithr . (l /\ r) /\ e)

-- Coswap

coswappwf :: Either a b -> Either b a
coswappwf = \x -> case x of Left  y -> inr y
		            Right z -> inl z

coswappff :: Either a b -> Either b a
coswappff = unpnt (curry (casesum snd (curry (inr . snd)) (curry (inl . snd))))


-- Streams

type Stream a = Mu (Const a :*: Id)

-- Tolist

tolist :: Stream a -> [a]
tolist = fix (\f -> \s -> inn (inr (fst (out s), f (snd (out s))))) 

-- Repeat

reppwf :: a -> Stream a
reppwf = fix (\s -> \x -> inn (x, s x))

reppff :: a -> Stream a
reppff = unpnt (fix . curry (curry (inn . (snd /\ app . (snd . fst /\ snd)))))

reppwh :: a -> Stream a
reppwh = hylo (_L :: Mu (Const a :*: Id)) g h
    where h = \x -> (x, x)
	  g = \y -> inn (fst y, snd y)

reppfh :: a -> Stream a
reppfh = hylo (_L :: Mu (Const a :*: Id)) g h
    where h = unpnt (curry (snd /\ snd))
	  g = unpnt (curry (inn . (fst . snd /\ snd . snd)))

rep'pwf :: a -> Stream a
rep'pwf = \x -> fix (\s -> inn (x,s))

rep'pff :: a -> Stream a
rep'pff = unpnt (curry (fix . curry (inn . (snd . fst /\ snd))))

-- Identity 

idpwf :: Stream a -> Stream a
idpwf = fix (\f -> \x -> inn (fst (out x), f (snd (out x))))

idpff :: Stream a -> Stream a
idpff = unpnt (fix . (curry (curry (inn . (fst . out . snd /\ app . (snd . fst /\ snd . out . snd))))))

idpwh :: Stream a -> Stream a
idpwh = hylo (_L :: Mu (Const (Stream a) :*: Id)) g h
    where h = \x -> (x, snd (out x))
	  g = \c -> inn (fst (out (fst c)), snd c)

idpfh :: Stream a -> Stream a
idpfh = hylo (_L :: Mu (Const (Stream a) :*: Id)) g h
    where h = unpnt (curry (snd /\ snd . out . snd))
	  g = unpnt (curry (inn . (fst . out . fst . snd /\ snd . snd)))

-- Length 

len :: [a] -> Int
len []     = 0
len (x:xs) = succ (len xs)

lenpwf :: [a] -> Int
lenpwf = fix (\f -> \x -> case (out x) of Left  y -> inn (inl _L)
                                          Right z -> inn (inr (f (snd z)))
             )

lenpff :: [a] -> Int
lenpff = unpnt (fix . curry (curry (casesum (out . snd) (curry (inn . inl . bang)) (curry (inn . inr . app . (snd . fst . fst /\ snd . snd))))))


lenpwh :: [a] -> Int
lenpwh = hylo (_L :: Mu (Const One :+: Id)) g h
    where h = \x -> case (out x) of Left  y -> inl _L
				    Right z -> inr (snd z)
	  g = \x -> case x of Left  y -> inn (inl _L)
			      Right z -> inn (inr z)
	  
lenpfh :: [a] -> Int
lenpfh = hylo (_L :: Mu (Const One :+: Id)) g h
    where h = unpnt (curry (casesum (out . snd) (curry (inl . bang)) (curry (inr . snd . snd))))
	  g = unpnt (curry (casesum (snd) (curry (inn . inl . bang)) (curry (inn . inr . snd))))
	  

lenpcf :: [a] -> Int
lenpcf = fix (\f -> \l -> if (null l) then 0
	                              else succ (f (tail l))
	     )  

-- Upto

gt :: (Int,Int) -> Bool
gt = uncurry (>)

upto :: (Int,Int) -> [Int]
upto (x,y) = if (x>y) then []
	              else x : upto (succ x, y)

uptopwf :: (Int,Int) -> [Int]
uptopwf = fix (\f -> \x -> case (out (gt x)) of Left  y -> inn (inl _L)
	                                        Right z -> inn (inr (fst x, f (inn (inr (fst x)), snd x)))
	      )

uptopwh :: (Int,Int) -> [Int]
uptopwh = hylo (_L :: Mu (Const One :+: Const (Int,Int) :*: Id)) g h
    where h = \x -> case (out (gt x)) of Left  y -> inl _L
	                                 Right z -> inr (x, (inn (inr (fst x)), snd x))
	  g = \x -> case x of Left  y -> inn (inl _L)
	                      Right z -> inn (inr (fst (fst z), snd z))

uptopfh :: (Int,Int) -> [Int]
uptopfh = hylo (_L :: Mu (Const One :+: Const (Int,Int) :*: Id)) g h
    where h = unpnt (curry (casesum (out . app . (pnt gt . bang /\ snd)) (curry (inl . bang)) (curry (inr . (snd . fst /\ (inn . inr . fst . snd . fst /\ snd . snd . fst))))))
	  g = unpnt (curry (casesum (snd) (curry (inn . inl . bang)) (curry (inn . inr . (fst . fst . snd /\ snd . snd)))))


-- Zip

zzip :: ([a],[b]) -> [(a,b)]
zzip ([],_)      = []
zzip (_,[])      = []
zzip (x:xs,y:ys) = (x,y) : zzip (xs,ys)

zippwf :: ([a],[b]) -> [(a,b)]
zippwf = fix (\f -> \x -> case (out (fst x)) 
	                  of Left  y -> inn (inl _L)
	                     Right z -> case (out (snd x))
	                                of Left  v -> inn (inl _L)
                                           Right w -> inn (inr ((fst z, fst w), f (snd z, snd w)))
	     )

zippwh :: ([a],[b]) -> [(a,b)]
zippwh = hylo (_L :: Mu(Const One :+: (Const One :+: (Const (a,[a]) :*: Const (b,[b])) :*: Id))) g h
    where h = \x -> case (out (fst x)) 
	            of Left  y -> inl _L
		       Right z -> case (out (snd x))
				  of Left  v -> inr (inl _L)
				     Right w -> inr (inr ((z, w), (snd z, snd w))) 
	  g = \x -> case x 
	            of Left  y -> inn (inl _L)
		       Right z -> case z
				  of Left  v -> inn (inl _L)
				     Right w -> inn (inr ((fst (fst (fst w)), fst (snd (fst w))), snd w))

zippfh :: ([a],[b]) -> [(a,b)]
zippfh = hylo (_L :: Mu(Const One :+: (Const One :+: (Const (a,[a]) :*: Const (b,[b])) :*: Id))) g h
    where h = unpnt (curry (casesum (out . fst . snd) (curry (inl . bang)) (curry (casesum (out . snd . snd . fst) (curry (inr . inl . bang)) (curry (inr . inr . ((snd . fst /\ snd) /\ (snd . snd . fst /\ snd . snd))))))))
	  g = unpnt (curry (casesum (snd) (curry (inn . inl . bang)) (curry (casesum (snd) (curry (inn . inl . bang)) (curry (inn . inr . ((fst . fst . fst . snd /\ fst . snd . fst . snd) /\ snd . snd)))))))

-- Map

mappwf :: (a -> b) -> [a] -> [b]
mappwf f = fix (\g -> \x -> case (out x)
	                    of Left  y -> inn (inl y)
	                       Right z -> inn (inr (f (fst z), g (snd z)))
	       )

mappwh :: (a -> b) -> [a] -> [b]
mappwh f = hylo (_L :: Mu (Const One :+: Const (a,[a]) :*: Id)) g h
    where h = \x -> case (out x)
	            of Left  y -> inl y
		       Right z -> inr (z, snd z)
	  g = \x -> case x
	            of Left  y -> inn (inl y)
		       Right z -> inn (inr (f (fst (fst z)), snd z))

mappfh :: (a -> b) -> [a] -> [b]
mappfh f = hylo (_L :: Mu (Const One :+: Const (a,[a]) :*: Id)) g h
    where h = unpnt (curry (casesum (out . snd) (curry (inl . snd)) (curry (inr . (snd /\ snd . snd)))))
	  g = unpnt (curry (casesum (snd) (curry (inn . inl . snd)) (curry (inn . inr . (app . (pnt f . bang /\ fst . fst . snd) /\ snd . snd)))))

-- Factorial

mult = uncurry (*)

fact :: Int -> Int
fact 0     = 1
fact (n+1) = mult (n+1, fact n)

factpwf :: Int -> Int
factpwf = fix (\f -> \x -> case (out x) of Left  z -> inn (inr (inn (inl _L)))
				           Right y -> mult (inn (inr y), f y)
	      )

factpwh :: Int -> Int
factpwh = hylo (_L :: Mu (Const One :+: Const Int :*: Id)) g h
    where h = \x -> case (out x) of Left  z -> inl _L
				    Right y -> inr (y, y)
	  g = \x -> case x of Left  z -> inn (inr (inn (inl _L)))
			      Right y -> mult (inn (inr (fst y)), snd y)
	  
factpfh :: Int -> Int
factpfh = hylo (_L :: Mu (Const One :+: Const Int :*: Id)) g h
    where h = unpnt (curry (casesum (out . snd) (curry (inl . bang)) (curry (inr . (snd /\ snd)))))
	  g = unpnt (curry (casesum (snd) (curry (inn . inr . inn . inl . bang)) (curry (app . (pnt mult . bang /\ (inn . inr . fst . snd /\ snd . snd))))))

-- Sum

plus = uncurry (+)

somapwf :: [Int] -> Int
somapwf = fix (\f -> \l -> case (out l) of Left  y -> inn (inl _L)
	                                   Right z -> plus (fst z, f (snd z))
	      )
somapwh :: [Int] -> Int
somapwh = hylo (_L :: Mu(Const One :+: Const (Int,[Int]) :*: Id)) g h
    where h = \l -> case (out l) of Left  y -> inl _L
				    Right z -> inr (z, snd z)
	  g = \x -> case x of Left  y -> inn (inl _L)
			      Right z -> plus (fst (fst z), snd z)

somapfh = hylo (_L :: Mu(Const One :+: Const (Int,[Int]) :*: Id)) g h
    where h = unpnt (curry (casesum (out . snd) (curry (inl . bang)) (curry (inr . (snd /\ snd . snd)))))
	  g = unpnt (curry (casesum (snd) (curry (inn . inl . bang)) (curry (app . (pnt plus . bang /\ (fst . fst . snd /\ snd . snd))))))

-- Fibonacci 

fib :: Int -> Int
fib 0     = 1
fib 1     = 1
fib (n+2) = plus (fib n, fib (n+1))

fibpwf :: Int -> Int
fibpwf = fix (\f -> \x -> case (out x) 
                          of Left  y -> inn (inr (inn (inl _L))) 
	                     Right z -> case (out z) 
			                of Left  v -> inn (inr (inn (inl _L)))
				           Right w -> plus (f w, f (inn (inr w)))
	     )

fibpwh :: Int -> Int
fibpwh = hylo (_L :: Mu (Const One :+: (Const One :+: (Id :*: Id)))) g h
    where h = \x -> case (out x) 
                    of Left  y -> inl _L 
		       Right z -> inr (case (out z) 
				       of Left  v -> inl _L
				          Right w -> inr (w, inn (inr w)))
	  g = \x -> case x 
                    of Left  y -> inn (inr (inn (inl _L)))
		       Right z -> case z
				  of Left  v -> inn (inr (inn (inl _L)))
				     Right w -> plus (fst w, snd w)

fibpfh :: Int -> Int
fibpfh = hylo (_L :: Mu (Const One :+: (Const One :+: (Id :*: Id)))) g h
    where h = unpnt (curry (casesum (out . snd) (curry (inl . bang)) (curry (inr . casesum (out . snd) (curry (inl . bang)) (curry (inr . (snd /\ inn . inr . snd)))))))
	  g = unpnt (curry (casesum (snd) (curry (inn . inr . inn . inl . bang)) (curry (casesum (snd) (curry (inn . inr . inn . inl . bang)) (curry (app . (pnt plus . bang /\ (fst . snd /\ snd . snd))))))))

